"""Unit test file for the OneWay class"""

import unittest
from unittest.mock import patch

from selenium.common.exceptions import NoSuchElementException

from flight_arbitrage.hidden_city import OneWay


class TestOneWay(unittest.TestCase):
    """Unit tests for the OneWay class"""

    def setUp(self):
        """

        :return:
        """
        self.flight = OneWay("start_airport", "end_airport", "date")

    def test_generate_browser_bad_assert(self):
        """

        :return:
        """
        with self.assertRaises(AssertionError):
            self.flight.generate_browser()

    def test_generate_browser_bad_exception(self):
        """

        :return:
        """
        with self.assertRaises(ValueError):
            with patch.object(self.flight, "browser", 0):
                self.flight.generate_browser()

    # https://stackoverflow.com/questions/60515794/mocking-instance-attributes
    @patch("flight_arbitrage.hidden_city.time")
    def test_generate_browser_good(self, mocked_time):
        """

        :param mocked_time:
        :return:
        """
        url = (
            "https://www.expedia.com/Flights-Search?trip=oneway&leg1="
            "from:start_airport,to:end_airport,"
            "departure:dateTANYT&passengers="
            "adults:1,children:0,seniors:0,infantinlap:Y"
            "&options=cabinclass%3Aeconomy&mode=search"
        )

        with patch.object(
            self.flight, "browser", unittest.mock.Mock()
        ) as mocked_browser:
            self.flight.generate_browser()

        mocked_browser.get.assert_called_once_with(url)
        mocked_time.sleep.assert_called_once_with(2)

    def test_retrieve_element_by_xpath(self):
        """

        :return:
        """
        base_object = unittest.mock.Mock()
        base_object.find_element_by_xpath.side_effect = [
            NoSuchElementException(),
            "element 1",
        ]

        results = []
        for _ in range(2):
            result = self.flight.retrieve_element_by_xpath(
                base_object, "context"
            )
            results.append(result)
            base_object.find_element_by_xpath.assert_called_with("context")

        self.assertTrue(None in results)
        self.assertTrue("element 1" in results)
        self.assertEqual(base_object.find_element_by_xpath.call_count, 2)

    def test_retrieve_elements_by_xpath(self):
        """

        :return:
        """
        base_object = unittest.mock.Mock()
        base_object.find_element_by_xpath.side_effect = [
            NoSuchElementException(),
            ["element 1", "element 2"],
        ]

        results = []
        for _ in range(2):
            result = self.flight.retrieve_elements_by_xpath(
                base_object, "context"
            )
            results.append(result)
            base_object.find_element_by_xpath.assert_called_with("context")

        self.assertTrue([] in results)
        self.assertTrue(["element 1", "element 2"] in results)
        self.assertEqual(base_object.find_element_by_xpath.call_count, 2)

    @patch("flight_arbitrage.hidden_city.time")
    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_elements_by_xpath")
    def test_cheapest_flight_bad_search(self, mocked_rxpath, mocked_time):
        """

        :param mocked_rxpath:
        :param mocked_time:
        :return:
        """
        mocked_rxpath.side_effect = [[], [], [], [], [], []]

        result = self.flight.cheapest_flight()

        self.assertEqual(result[0], -1)
        self.assertTrue(isinstance(result[0], float))
        self.assertTrue(isinstance(result[1], dict))

        mocked_rxpath.assert_called_with(
            None, '//li[@data-test-id="offer-listing"]'
        )
        self.assertEqual(mocked_rxpath.call_count, 4)
        mocked_time.sleep.assert_called_with(1)
        self.assertEqual(mocked_time.sleep.call_count, 3)

    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_elements_by_xpath")
    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_element_by_xpath")
    def test_cheapest_flight_good_search_empty(
        self, mocked_rxpath_element, mocked_rxpath_elements
    ):
        """

        :param mocked_rxpath_element:
        :param mocked_rxpath_elements:
        :return:
        """
        departure_time = '//span[@data-test-id="departure-time"]'
        price = '//span[@class="uitk-lockup-price"]'

        mocked_rxpath_elements.side_effect = [["element 1", "element 2"]]
        mocked_rxpath_element.side_effect = ["", "search_departure", ""]

        result = self.flight.cheapest_flight()

        self.assertEqual(result[0], -1)
        self.assertTrue(isinstance(result[0], float))
        self.assertTrue(isinstance(result[1], dict))

        mocked_rxpath_elements.assert_called_once_with(
            None, '//li[@data-test-id="offer-listing"]'
        )
        calls = [
            unittest.mock.call("element 1", "." + departure_time),
            unittest.mock.call("element 2", "." + departure_time),
            unittest.mock.call("element 2", "." + price),
        ]
        mocked_rxpath_element.assert_has_calls(calls, any_order=False)
        self.assertEqual(mocked_rxpath_element.call_count, len(calls))

    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_elements_by_xpath")
    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_element_by_xpath")
    def test_cheapest_flight_good_search_full(
        self, mocked_rxpath_element, mocked_rxpath_elements
    ):
        """

        :param mocked_rxpath_element:
        :param mocked_rxpath_elements:
        :return:
        """
        departure_time = '//span[@data-test-id="departure-time"]'
        price = '//span[@class="uitk-lockup-price"]'

        mocked_departure = unittest.mock.MagicMock()
        mocked_departure.text.split().__getitem__.return_value = "location"
        mocked_price = unittest.mock.MagicMock()
        mocked_price.text.__getitem__.return_value = "58"
        mocked_rxpath_elements.side_effect = [
            ["element 1", "element 2", "element 3"]
        ]
        mocked_rxpath_element.side_effect = [
            "",
            "search_departure",
            "",
            mocked_departure,
            mocked_price,
        ]

        result = self.flight.cheapest_flight()

        self.assertEqual(result[0], 58)
        self.assertTrue(isinstance(result[0], float))
        self.assertTrue(isinstance(result[1], dict))
        self.assertEqual(result[1], {"location": {58}})

        mocked_rxpath_elements.assert_called_once_with(
            None, '//li[@data-test-id="offer-listing"]'
        )
        calls = [
            unittest.mock.call("element 1", "." + departure_time),
            unittest.mock.call("element 2", "." + departure_time),
            unittest.mock.call("element 2", "." + price),
            unittest.mock.call("element 3", "." + departure_time),
            unittest.mock.call("element 3", "." + price),
        ]
        mocked_rxpath_element.assert_has_calls(calls, any_order=False)
        self.assertEqual(mocked_rxpath_element.call_count, len(calls))

    @patch("flight_arbitrage.hidden_city.tqdm")
    @patch("flight_arbitrage.hidden_city.OneWay.cheapest_flight")
    @patch("flight_arbitrage.hidden_city.OneWay.airports_to_search")
    @patch("flight_arbitrage.hidden_city.OneWay.generate_browser")
    @patch("flight_arbitrage.hidden_city.OneWay.open_browser")
    def test_find_arbitrage_exception(
        self,
        mocked_open,
        mocked_generate,
        mocked_search,
        mocked_cheapest,
        mocked_tqdm,
    ):
        """

        :param mocked_open:
        :param mocked_generate:
        :param mocked_search:
        :param mocked_cheapest:
        :param mocked_tqdm:
        :return:
        """
        mocked_search.return_value = ["airport 1", "airport 2"]
        mocked_cheapest.return_value = [
            100,
            {"location": {58}},
        ]  # base, departure_dict
        mocked_tqdm.return_value = mocked_search.return_value

        with patch.object(
            self.flight, "browser", unittest.mock.Mock()
        ) as mocked_browser:
            mocked_browser.get.side_effect = Exception()

            result = self.flight.find_arbitrage()

        mocked_open.assert_called_once_with(
            web_browser="firefox", driver="", headless=False
        )
        mocked_generate.assert_called_once_with()
        mocked_search.assert_called_once_with(
            override=False, override_filename="airports.txt"
        )
        mocked_cheapest.assert_called_once_with()
        mocked_tqdm.assert_called_once_with(["airport 1", "airport 2"])
        self.assertEqual(result, [])

    @patch("flight_arbitrage.hidden_city.re")
    @patch("flight_arbitrage.hidden_city.time")
    @patch("flight_arbitrage.hidden_city.tqdm")
    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_elements_by_xpath")
    @patch("flight_arbitrage.hidden_city.OneWay.retrieve_element_by_xpath")
    @patch("flight_arbitrage.hidden_city.OneWay.cheapest_flight")
    @patch("flight_arbitrage.hidden_city.OneWay.airports_to_search")
    @patch("flight_arbitrage.hidden_city.OneWay.generate_browser")
    @patch("flight_arbitrage.hidden_city.OneWay.open_browser")
    def test_find_arbitrage(
        self,
        mocked_open,
        mocked_generate,
        mocked_search,
        mocked_cheapest,
        mocked_rxpath_element,
        mocked_rxpath_elements,
        mocked_tqdm,
        mocked_time,
        mocked_re,
    ):
        """

        :param mocked_open:
        :param mocked_generate:
        :param mocked_search:
        :param mocked_cheapest:
        :param mocked_rxpath_element:
        :param mocked_rxpath_elements:
        :param mocked_tqdm:
        :param mocked_time:
        :param mocked_re:
        :return:
        """
        departure_time = '//span[@data-test-id="departure-time"]'
        price = '//span[@class="uitk-lockup-price"]'
        layovers = '//div[@data-test-id="layovers"]'

        mocked_search.return_value = [
            "start_airport",
            "airport_2",
            "airport_3",
        ]
        mocked_cheapest.return_value = [
            100,
            {"location": {60}, "olympics": {}},
        ]  # base, departure_dict
        mocked_tqdm.return_value = mocked_search.return_value

        mocked_new_searches = unittest.mock.MagicMock()
        mocked_new_searches.text = "new search"
        mocked_re.findall.return_value = ["stop 1", "stop 2", "end_airport"]
        mocked_departure = unittest.mock.MagicMock()  # can handle slicing
        mocked_departure.text.split().__getitem__.side_effect = [
            "location",
            "olympics",
        ]
        mocked_price = unittest.mock.MagicMock()  # can handle slicing
        mocked_price.text.__getitem__.side_effect = ["99", "58", "58"]
        mocked_rxpath_elements.side_effect = [
            [],
            ["element 1", "element 2", "element 3", "element 4", "element 5"],
            ["element 6"],
        ]
        mocked_rxpath_element.side_effect = [
            "",
            mocked_new_searches,
            "",
            mocked_new_searches,
            mocked_price,
            "",
            mocked_new_searches,
            mocked_price,
            mocked_departure,
            mocked_new_searches,
            mocked_price,
            mocked_departure,
            "",
        ]
        url = (
            "https://www.expedia.com/Flights-Search?trip=oneway&"
            "leg1=from:start_airport,to:airport_2,"
            "departure:dateTANYT&passengers=adults:1,children:0,seniors:0"
            ",infantinlap:Y"
            "&options=cabinclass%3Aeconomy&mode=search"
        )

        with patch.object(
            self.flight, "browser", unittest.mock.Mock()
        ) as mocked_browser:
            result = self.flight.find_arbitrage()

        mocked_browser.get.assert_any_call(url)
        self.assertEqual(mocked_browser.get.call_count, 2)

        expected_result = [
            {
                "airport destination": "end_airport",
                "airport source": "start_airport",
                "base price": 100,
                "this ticket price": 58.0,
                "eval price": 60,
                "this destination": "airport_2",
                "savings": 2.0,
            },
            {
                "airport destination": "end_airport",
                "airport source": "start_airport",
                "base price": 100,
                "this ticket price": 58.0,
                "eval price": 100,
                "this destination": "airport_2",
                "savings": 42.0,
            },
        ]
        self.assertEqual(result, expected_result)

        mocked_open.assert_called_once_with(
            web_browser="firefox", driver="", headless=False
        )
        mocked_generate.assert_called_once_with()
        mocked_search.assert_called_once_with(
            override=False, override_filename="airports.txt"
        )
        mocked_cheapest.assert_called_once_with()
        mocked_tqdm.assert_called_once_with(
            ["start_airport", "airport_2", "airport_3"]
        )

        calls = [
            unittest.mock.call(2),
            unittest.mock.call(1),
            unittest.mock.call(2),
            unittest.mock.call(2),
            unittest.mock.call(2),
        ]
        mocked_time.sleep.assert_has_calls(calls, any_order=False)
        self.assertTrue(mocked_time.sleep.call_count, len(calls))

        mocked_re.findall.assert_any_call(r"\(.*?\)", "new search")
        self.assertTrue(mocked_re.findall.call_count, 4)

        mocked_rxpath_elements.assert_any_call(
            mocked_browser, '//li[@data-test-id="offer-listing"]'
        )
        self.assertTrue(mocked_rxpath_elements.call_count, 3)
        calls = [
            unittest.mock.call("element 1", "." + layovers),
            unittest.mock.call("element 2", "." + layovers),
            unittest.mock.call("element 2", "." + price),
            unittest.mock.call("element 3", "." + layovers),
            unittest.mock.call("element 3", "." + price),
            unittest.mock.call("element 3", "." + departure_time),
            unittest.mock.call("element 4", "." + layovers),
            unittest.mock.call("element 4", "." + price),
            unittest.mock.call("element 4", "." + departure_time),
            unittest.mock.call("element 5", "." + layovers),
            unittest.mock.call("element 5", "." + price),
            unittest.mock.call("element 5", "." + departure_time),
            unittest.mock.call("element 6", "." + layovers),
        ]
        mocked_rxpath_element.assert_has_calls(calls, any_order=False)
        self.assertTrue(mocked_rxpath_element.call_count, len(calls))

        mocked_browser.quit.assert_called_once_with()


if __name__ == "__main__":

    unittest.main()



"""Unit test file for the Flight class"""

import unittest
from unittest.mock import patch

from flight_arbitrage.flight import Flight


class TestFlight(unittest.TestCase):
    """Unit tests for the Flight class"""

    def setUp(self):
        """
        Create the initial class object for use in each unit test

        :return: nothing
        """
        self.flight = Flight("start_airport", "end_airport", "date")

    @patch("flight_arbitrage.flight.ChromeOptions")
    @patch("flight_arbitrage.flight.webdriver")
    def test_open_chrome(self, mocked_webdriver, mocked_options):
        """
        Opening a chrome driver using selenium

        :param mocked_webdriver: a mocked selenium webdriver object
        :param mocked_options: a mocked function for Chrome driver options
        :return: nothing
        """
        # headless is False
        path = "path"

        self.flight.open_chrome(driver=path)

        mocked_webdriver.Chrome.assert_called_once_with(executable_path="path")

        # headless is True
        headless = True

        self.flight.open_chrome(driver=path, headless=headless)

        mocked_options.add_argument("--headless")
        mocked_options.add_argument.assert_called_once_with("--headless")
        self.assertEqual(mocked_webdriver.Chrome.call_count, 2)
        mocked_webdriver.Chrome.assert_called_with(
            executable_path="path", options=mocked_options()
        )

    @patch("flight_arbitrage.flight.FirefoxOptions")
    @patch("flight_arbitrage.flight.webdriver")
    def test_open_firefox(self, mocked_webdriver, mocked_options):
        """

        :param mocked_webdriver:
        :param mocked_options:
        :return:
        """
        # headless is False
        self.flight.open_firefox()

        mocked_webdriver.Firefox.assert_called_once_with()

        # headless is True
        headless = True

        self.flight.open_firefox(headless=headless)

        mocked_options.add_argument("--headless")
        mocked_options.add_argument.assert_called_once_with("--headless")
        self.assertEqual(mocked_webdriver.Firefox.call_count, 2)
        mocked_webdriver.Firefox.assert_called_with(options=mocked_options())

    @patch("flight_arbitrage.flight.webdriver")
    def test_open_safari(self, mocked_webdriver):
        """

        :param mocked_webdriver:
        :return:
        """
        # headless is False
        path = "path"

        self.flight.open_safari(driver=path)

        mocked_webdriver.Safari.assert_called_once_with(executable_path="path")

        # headless is True
        headless = True

        self.flight.open_safari(driver=path, headless=headless)

        self.assertEqual(mocked_webdriver.Safari.call_count, 2)
        mocked_webdriver.Safari.assert_called_with(executable_path="path")

    @patch("flight_arbitrage.flight.webdriver")
    def test_open_edge(self, mocked_webdriver):
        """

        :param mocked_webdriver:
        :return:
        """
        # headless is False
        path = "path"

        self.flight.open_edge(driver=path)

        mocked_webdriver.Edge.assert_called_once_with(executable_path="path")

        # headless is True
        headless = True

        self.flight.open_edge(driver=path, headless=headless)

        self.assertEqual(mocked_webdriver.Edge.call_count, 2)
        mocked_webdriver.Edge.assert_called_with(executable_path="path")

    def test_open_browser_bad_exception(self):
        """

        :return:
        """
        # bad browser
        with self.assertRaises(Exception):
            self.flight.open_browser(web_browser="browser")

    @patch("flight_arbitrage.flight.Flight.open_edge")
    @patch("flight_arbitrage.flight.Flight.open_safari")
    @patch("flight_arbitrage.flight.Flight.open_firefox")
    @patch("flight_arbitrage.flight.Flight.open_chrome")
    @patch("flight_arbitrage.flight.time")
    def test_open_browser_good(
        self,
        mocked_time,
        mocked_chrome,
        mocked_firefox,
        mocked_safari,
        mocked_edge,
    ):
        """

        :param mocked_time:
        :param mocked_chrome:
        :param mocked_firefox:
        :param mocked_safari:
        :param mocked_edge:
        :return:
        """
        # chrome
        self.flight.open_browser(
            web_browser="chrome", driver="path", headless=False
        )

        mocked_time.sleep.assert_called_once_with(2)
        mocked_chrome.assert_called_once_with(driver="path", headless=False)

        # firefox
        self.flight.open_browser(web_browser="firefox", headless=False)

        mocked_time.sleep.assert_called_with(2)
        self.assertEqual(mocked_time.sleep.call_count, 2)
        mocked_firefox.assert_called_once_with(headless=False)

        # safari
        self.flight.open_browser(
            web_browser="safari", driver="path", headless=False
        )

        mocked_time.sleep.assert_called_with(2)
        self.assertEqual(mocked_time.sleep.call_count, 3)
        mocked_safari.assert_called_once_with(driver="path", headless=False)

        # edge
        self.flight.open_browser(
            web_browser="edge", driver="path", headless=False
        )

        mocked_time.sleep.assert_called_with(2)
        self.assertEqual(mocked_time.sleep.call_count, 4)
        mocked_edge.assert_called_once_with(driver="path", headless=False)

    @patch("flight_arbitrage.flight.requests.get")
    def test_airports_to_search_bad_request_status(self, mocked_get):
        """

        :param mocked_get:
        :return:
        """
        mocked_get.return_value.status_code = 201

        with self.assertRaises(ValueError):
            self.flight.airports_to_search()

        self.assertEqual(mocked_get.call_count, 1)

    @patch("flight_arbitrage.flight.requests.get")
    def test_airports_to_search_bad_request_exception(self, mocked_get):
        """

        :param mocked_get:
        :return:
        """
        mocked_get.side_effect = Exception()

        with self.assertRaises(ValueError):
            self.flight.airports_to_search()

        self.assertEqual(mocked_get.call_count, 1)

    @patch("flight_arbitrage.flight.BeautifulSoup")
    @patch("flight_arbitrage.flight.requests.get")
    def test_airports_to_search_good_request(self, mocked_get, mocked_bs4):
        """

        :param mocked_get:
        :param mocked_bs4:
        :return:
        """
        url = (
            "https://en.wikipedia.org/wiki/List_of_the_"
            "busiest_airports_in_the_United_States"
        )
        mocked_get.return_value.status_code = 200
        mocked_get.return_value.text = "airport list"

        mocked_rows = unittest.mock.MagicMock()
        mocked_row = unittest.mock.MagicMock()

        mocked_cells = unittest.mock.MagicMock()
        mocked_cells.__getitem__.return_value.string.strip.side_effect = [
            "major airport",
            "minor airport",
        ]

        mocked_row.findChildren.return_value = mocked_cells
        mocked_rows.findChildren.return_value = [
            unittest.mock.MagicMock(),
            mocked_row,
        ]

        mocked_bs4.return_value.findChildren.side_effect = [
            [mocked_rows, mocked_rows]
        ]

        result = self.flight.airports_to_search()

        mocked_get.assert_called_once_with(url)

        mocked_bs4.assert_called_once_with("airport list", features="lxml")
        mocked_bs4.return_value.findChildren.assert_called_once_with("table")

        mocked_rows.findChildren.assert_any_call(["tr"])
        self.assertTrue(mocked_rows.findChildren.call_count, 2)
        mocked_row.findChildren.assert_any_call("td")
        self.assertTrue(mocked_row.findChildren.call_count, 2)

        self.assertEqual(result, ["major airport", "minor airport"])

    @patch("builtins.open")
    def test_airports_to_search_custom_airports(self, mocked_open):
        """

        :param mocked_open:
        :return:
        """
        mocked_open.return_value.__enter__.return_value = [
            " airport1",
            "airport2 ",
        ]

        result = self.flight.airports_to_search(
            override=True, override_filename="filename"
        )
        expected_result = ["airport1", "airport2"]

        mocked_open.assert_called_once_with("filename", "r")
        self.assertTrue(result, expected_result)

        result = self.flight.airports_to_search(override=True)
        expected_result = ["airport1", "airport2"]

        mocked_open.assert_called_with("airports.txt", "r")
        self.assertEqual(mocked_open.call_count, 2)
        self.assertEqual(result, expected_result)


if __name__ == "__main__":

    unittest.main()



"""Find arbitrage in plane ticket prices"""

import re
import time
from typing import Tuple, DefaultDict, List, Union, Optional
from collections import defaultdict

from tqdm import tqdm  # type: ignore
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException

from flight_arbitrage.flight import Flight

ElementType = Optional[
    Union[
        webdriver.remote.webelement.WebElement,
        webdriver.firefox.webelement.FirefoxWebElement,
    ]
]
ElementsType = Union[
    List[webdriver.remote.webelement.WebElement],
    List[webdriver.firefox.webelement.FirefoxWebElement],
]


class OneWay(Flight):
    """Find arbitrage opportunities in one-way flights"""

    def generate_browser(self) -> None:
        """
        Opens a browser and goes to the expedia flights website

        :return: nothing
        """
        url = (
            f"https://www.expedia.com/Flights-Search?trip=oneway&leg1="
            f"from:{self.leaving_from},to:{self.going_to},"
            f"departure:{self.date}"
            f"TANYT&passengers=adults:1,children:0,seniors:0,infantinlap:Y"
            f"&options=cabinclass%3Aeconomy&mode=search"
        )

        assert (
            self.browser is not None
        ), "browser variable is the wrong data type"

        try:
            self.browser.get(url)
        except Exception as error:
            raise ValueError(
                f"error opening up browser with error: {error}"
            ) from error

        time.sleep(2)

    @staticmethod
    def retrieve_element_by_xpath(base_object, context: str) -> ElementType:
        """
        Searches the site for a specific string -- context

        :param base_object: html parent tag object
        :param context: string to search
        :return: xpath object from selenium
        """
        searching: ElementType = None

        try:
            searching = base_object.find_element_by_xpath(context)
        except NoSuchElementException:
            pass

        return searching

    @staticmethod
    def empty_list() -> List[webdriver.firefox.webelement.FirefoxWebElement]:
        """
        Creates an empty list due to
            mypy issue: https://github.com/python/mypy/issues/6463
        The type in function 'retrieve_elements_by_xpath'
            variable 'searching' defaults to a firefox empty list

        :return: a default empty list to help work around a mypy bug
        """
        return []

    def retrieve_elements_by_xpath(
        self, base_object, context: str
    ) -> ElementsType:
        """
        Searches the site for a specific string -- context

        :param base_object: html parent tag object
        :param context: string to search
        :return: list of xpath object from selenium
        """
        # uses an extra function due to open mypy issue:
        #   https://github.com/python/mypy/issues/6463
        searching: ElementsType = self.empty_list()

        try:
            searching = base_object.find_element_by_xpath(context)
        except NoSuchElementException:
            pass

        return searching

    def cheapest_flight(
        self, tries: int = 3
    ) -> Tuple[float, DefaultDict[str, set]]:
        """


        :param tries:
        :return:
        """
        try_count = 0
        search = self.retrieve_elements_by_xpath(self.browser, self.offerings)

        departure_dict: DefaultDict[str, set] = defaultdict(set)
        while not search and try_count < tries:
            search = self.retrieve_elements_by_xpath(
                self.browser, self.offerings
            )

            try_count += 1
            time.sleep(1)

        if not search and tries == 3:
            return -1.0, departure_dict

        if search:
            for search_object in search:
                search_departure = self.retrieve_element_by_xpath(
                    search_object, "." + self.departure_time
                )
                if not search_departure:
                    continue

                search_price = self.retrieve_element_by_xpath(
                    search_object, "." + self.price
                )
                if not search_price:
                    continue

                return_price = float(
                    search_price.text[1:].replace(",", "").replace(".", "")
                )
                departure_location = search_departure.text.split("-")[
                    0
                ].strip()
                departure_dict[departure_location].add(return_price)

                return return_price, departure_dict

        return -1.0, departure_dict

    def find_arbitrage(
        self,
        override: bool = False,
        override_filename: str = "airports.txt",
        web_browser: str = "firefox",
        driver: str = "",
        headless: bool = False,
        tries: int = 3,
    ) -> list:
        """


        :param override:
        :param override_filename:
        :param web_browser:
        :param driver:
        :param headless:
        :param tries:
        :return:

        >>> arbitrage = OneWay('JFK', 'SLC', '07/10/2021')
        >>> a = arbitrage.find_arbitrage(headless=True)
        >>> print(f'Is there an arbitrage opportunity: {len(a) > 0}')
        >>> for d in a:
        >>>     print(d)
        """
        self.open_browser(
            web_browser=web_browser, driver=driver, headless=headless
        )
        self.generate_browser()

        # have to have the below assert --> related to mypy issue:
        #   https://github.com/python/mypy/issues/5528
        assert (
            self.browser is not None
        ), "browser variable is the wrong data type"

        airports = self.airports_to_search(
            override=override, override_filename=override_filename
        )
        base, departure_dict = self.cheapest_flight()
        arbs = []

        for airport in tqdm(airports):
            if airport == self.leaving_from:
                continue

            url = (
                f"https://www.expedia.com/Flights-Search?trip="
                f"oneway&leg1=from:{self.leaving_from},"
                f"to:{airport},departure:{self.date}"
                f"TANYT&passengers=adults:1,children:0,seniors:0,"
                f"infantinlap:Y&options=cabinclass%3Aeconomy&mode=search"
            )

            try:
                self.browser.get(url)
            except Exception as error:
                print(f"error opening up browser with error: {error}")
                return []

            time.sleep(2)

            try_count = 0
            search = self.retrieve_elements_by_xpath(
                self.browser, self.offerings
            )

            while not search and try_count < tries:
                search = self.retrieve_elements_by_xpath(
                    self.browser, self.offerings
                )

                try_count += 1
                time.sleep(1)

            lowest_ticket = []
            bad_count = 0
            for search_object in search:
                new_searches = self.retrieve_element_by_xpath(
                    search_object, "." + self.layovers
                )
                if not new_searches:
                    bad_count += 1
                    continue
                find_all = re.findall(r"\(.*?\)", new_searches.text)
                stops = [location.strip("()") for location in find_all]

                price_found = self.retrieve_element_by_xpath(
                    search_object, "." + self.price
                )
                if not price_found:
                    bad_count += 1
                    continue

                ticket_price = float(
                    price_found.text[1:].replace(",", "").replace(".", "")
                )
                if ticket_price < base and self.going_to in stops:
                    print("arbitrage with destination:", airport)

                    departure = self.retrieve_element_by_xpath(
                        search_object, "." + self.departure_time
                    )
                    if not departure:
                        continue
                    departure_value = departure.text.split("-")[0].strip()

                    evaluation_price = base
                    if not departure_dict[departure_value]:
                        savings = base - ticket_price
                    else:
                        evaluation_price = min(departure_dict[departure_value])
                        savings = (
                            min(departure_dict[departure_value]) - ticket_price
                        )
                    arbs.append(
                        {
                            "airport destination": self.going_to,
                            "airport source": self.leaving_from,
                            "base price": base,
                            "this ticket price": ticket_price,
                            "eval price": evaluation_price,
                            "this destination": airport,
                            "savings": savings,
                        }
                    )
                lowest_ticket.append(ticket_price)

            if search and len(search) != bad_count:
                print(
                    f"done with airport: {airport} | "
                    f"cheapest ticket with layovers: {min(lowest_ticket)}"
                )
            else:
                print(
                    f"done with airport: {airport} | "
                    f"no available flights from {self.leaving_from}"
                    f" to {airport}"
                )

            time.sleep(2)

        self.browser.quit()

        return arbs



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['flight_arbitrage']

package_data = \
{'': ['*']}

setup(name='flight_arbitrage',
      version='0.7.1',
      description='Finds arbitrage in plane ticket prices',
      author='Jordan Rinder',
      author_email='jordan.rinder@gmail.com',
      url='https://github.com/jrinder42/Flight_Arbitrage',
      packages=packages,
      package_data=package_data,
      python_requires='>=3.7',
     )



"""Creates flight data scraping object"""

import time
from typing import List, Union

import requests
from bs4 import BeautifulSoup  # type: ignore
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions


class Flight:
    """Handle browser scraping"""

    def __init__(self, leaving_from: str, going_to: str, date: str) -> None:
        """
        Flight constructor

        :param leaving_from: airport where the flight originates
        :param going_to: airport that is the flight destination
        :param date: date of flight
        """
        self.leaving_from = leaving_from
        self.going_to = going_to
        self.date = date

        self.browser: Union[
            webdriver.Chrome,
            webdriver.Firefox,
            webdriver.Safari,
            webdriver.Edge,
            None,
        ] = None

        self.price = '//span[@class="uitk-lockup-price"]'
        self.layovers = '//div[@data-test-id="layovers"]'
        self.offerings = '//li[@data-test-id="offer-listing"]'
        self.departure_time = '//span[@data-test-id="departure-time"]'

    def open_chrome(self, driver: str = "", headless: bool = False) -> None:
        """
        Open the chrome browser

        :param driver: chrome driver file path
        :param headless: headless browser mode
        :return: nothing
        """
        if headless:
            options = ChromeOptions()
            options.add_argument("--headless")
            self.browser = webdriver.Chrome(
                executable_path=driver, options=options
            )
        else:
            self.browser = webdriver.Chrome(executable_path=driver)

    def open_firefox(self, headless: bool = False) -> None:
        """
        Open the firefox browser

        :param headless: headless browser mode
        :return: nothing
        """
        if headless:
            options = FirefoxOptions()
            options.add_argument("--headless")
            self.browser = webdriver.Firefox(options=options)
        else:
            self.browser = webdriver.Firefox()

    def open_safari(self, driver: str = "", headless: bool = False) -> None:
        """
        Open the safari browser

        :param driver: safari driver file path
        :param headless: headless browser mode
        :return: nothing
        """
        if headless:
            print(
                "sorry, headless is not available with "
                "safari in selenium at this moment"
            )
        self.browser = webdriver.Safari(executable_path=driver)

    def open_edge(self, driver: str = "", headless: bool = False) -> None:
        """
        Open the edge browser

        :param driver: edge driver file path
        :param headless: headless browser mode
        :return: nothing
        """
        if headless:
            print(
                "sorry, headless is not available with edge in "
                "selenium at this moment (well it kinda is)"
            )
        self.browser = webdriver.Edge(executable_path=driver)

    def open_browser(
        self,
        web_browser: str = "firefox",
        driver: str = "",
        headless: bool = False,
    ) -> None:
        """
        Opens a browser based on user-defined parameters

        :param web_browser: web browser to open
        :param driver: web browser driver file path
        :param headless: headless browser mode
        :return: nothing
        """
        web_browser = web_browser.lower()
        if web_browser == "chrome":
            self.open_chrome(driver=driver, headless=headless)
        elif web_browser == "edge":
            self.open_edge(driver=driver, headless=headless)
        elif web_browser == "firefox":
            self.open_firefox(headless=headless)
        elif web_browser == "safari":
            self.open_safari(driver=driver, headless=headless)
        else:
            raise Exception(f"web browser {web_browser} is not available")

        time.sleep(2)

    @staticmethod
    def airports_to_search(
        override: bool = False, override_filename: str = "airports.txt"
    ) -> List[str]:
        """
        Creates list of airports to iterate through

        :param override: indicating whether the user will use a custom file
        :param override_filename: the path/filename of the custom airport list
        :return: a list of airports to iterate through
        """
        if override:
            with open(override_filename, "r") as file:
                airports = [line.strip() for line in file]
            return airports

        url = (
            "https://en.wikipedia.org/wiki/List_of_the_busiest_"
            "airports_in_the_United_States"
        )
        try:
            response = requests.get(url)
            if response.status_code == 200:
                response_text = response.text
            else:
                raise ValueError(
                    f"did not return the correct response: "
                    f"{response.status_code}"
                )
        except Exception as error:
            raise ValueError(
                f"something went wrong with the url pull: {error}"
            ) from error

        airports = []
        soup = BeautifulSoup(response_text, features="lxml")
        tables = soup.findChildren("table")
        for airport_table in [tables[0], tables[1]]:
            rows = airport_table.findChildren(["tr"])
            for i, row in enumerate(rows):
                if i == 0:
                    continue

                cells = row.findChildren("td")
                airports.append(cells[2].string.strip("\n"))

        return airports
